(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 25-Apr-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: HDRCubeMap.cpp
//
// Desc: DirectX window application created by the DirectX AppWizard
//-----------------------------------------------------------------------------

unit HDRCubeMapUnit;

interface

uses
  Windows, Messages,
  Direct3D9, D3DX9, dxerr9,
  D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil,
  Resource;

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders



//-----------------------------------------------------------------------------
// Defines, and constants
//-----------------------------------------------------------------------------

const
  ENVMAPSIZE = 256;
// Currently, 4 is the only number of lights supported.
  NUM_LIGHTS = 4;
  LIGHTMESH_RADIUS = 0.2;

const
  HELPTEXTCOLOR: TD3DXColor = (r: 0.0; g:1.0; b:0.3; a:1.0);

var
  g_aVertDecl: array[0..3] of TD3DVertexElement9 =
  (
    (Stream: 0; Offset:  0; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_POSITION; UsageIndex: 0),
    (Stream: 0; Offset: 12; _Type: D3DDECLTYPE_FLOAT3; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_NORMAL; UsageIndex: 0),
    (Stream: 0; Offset: 24; _Type: D3DDECLTYPE_FLOAT2; Method: D3DDECLMETHOD_DEFAULT; Usage: D3DDECLUSAGE_TEXCOORD; UsageIndex: 0),
    ({D3DDECL_END})
  );



  g_atszMeshFileName: array[0..2] of PChar = (
    'teapot.x',
    'skullocc.x',
    'car.x'
  );

const
  //#define NUM_MESHES ( sizeof(g_atszMeshFileName) / sizeof(g_atszMeshFileName[0]) )
  NUM_MESHES = SizeOf(g_atszMeshFileName) div SizeOf(g_atszMeshFileName[0]);

type
  TOrbitData = record
    tszXFile: PChar;        // X file
    vAxis:    TD3DVector;   // Axis of rotation
    fRadius:  Single;       // Orbit radius
    fSpeed:   Single;       // Orbit speed in radians per second
  end;


const
  // Mesh file to use for orbit objects
  // These don't have to use the same mesh file.
  g_OrbitData: array[0..1] of TOrbitData = (
    (tszXFile: 'airplane 2.x'; vAxis:(x:-1.0; y:0.0; z:0.0); fRadius: 2.0; fSpeed: {D3DX_PI}3.141592654 * 1.0),
    (tszXFile: 'airplane 2.x'; vAxis:(x: 0.3; y:1.0; z:0.3); fRadius: 2.5; fSpeed: {D3DX_PI}3.141592654 / 2.0)
  );


type
  //-----------------------------------------------------------------------------
  // Name: struct CObj
  // Desc: Encapsulate an object in the scene with its world transformation
  //       matrix.
  //-----------------------------------------------------------------------------
  CObj = class 
  private
    m_mWorld: TD3DXMatrixA16;  // World transformation
    m_Mesh:   CD3DMesh;        // Mesh object

  public
    constructor Create; { D3DXMatrixIdentity( &m_mWorld ); }
    destructor Destroy; override;
    function WorldCenterAndScaleToRadius(fRadius: Single): HRESULT;
  end;


  //-----------------------------------------------------------------------------
  // Name: class COrbit
  // Desc: Encapsulate an orbit object in the scene with related data
  //-----------------------------------------------------------------------------
  COrbit = class(CObj)
  public
    m_vAxis:   TD3DXVector3; // orbit axis
    m_fRadius: Single;       // orbit radius
    m_fSpeed:  Single;       // Speed, angle in radian per second

  public
    constructor Create; // COrbit() : m_vAxis( 0.0f, 1.0f, 0.0f ), m_fRadius(1.0f), m_fSpeed( D3DX_PI ) { }
    procedure SetOrbit(const vAxis: TD3DXVector3; fRadius, fSpeed: Single);
    procedure Orbit(fElapsedTime: Single);
  end;


type
  TLight = record
    vPos: TD3DXVector4;        // Position in world space
    vMoveDir: TD3DXVector4;    // Direction in which it moves
    fMoveDist: Single;         // Maximum distance it can move
    mWorld: TD3DXMatrixA16;    // World transform matrix for the light before animation
    mWorking: TD3DXMatrixA16;  // Working matrix (world transform after animation)
  end;


type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont: ID3DXFont;                     // Font for drawing text
    m_pFontSmall: ID3DXFont;                // Font for drawing text
    m_Camera: CModelViewerCamera;           // View camera
    m_aLights: array[0..NUM_LIGHTS-1] of TLight; // Parameters of light objects
    m_vLightIntensity: TD3DXVector4;        // Light intensity
    m_pEffect: ID3DXEffect;                 // Effect for the sample
    m_EnvMesh: array[0..NUM_MESHES-1] of CObj; // Mesh to receive environment mapping
    m_nCurrMesh: Integer;                   // Index of the element in m_EnvMesh we are currently displaying
    m_pRoom: CD3DMesh;                      // Mesh representing room (wall, floor, ceiling)
    m_Orbit: array[0.. SizeOf(g_OrbitData) div SizeOf(g_OrbitData[0])-1] of COrbit; // Orbit meshes
    m_pLightMesh: CD3DMesh;                 // Mesh for the light object
    m_pVertDecl: IDirect3DVertexDeclaration9; // Vertex decl for the sample
    m_pCubeMapFp: IDirect3DCubeTexture9;    // Floating point format cube map
    m_pCubeMap32: IDirect3DCubeTexture9;    // 32-bit cube map (for fallback)
    m_pDepthCube: IDirect3DSurface9;        // Depth-stencil buffer for rendering to cube texture

    m_hTechRenderScene: TD3DXHandle;        // Technique handle for RenderScene
    m_hTechRenderHDREnvMap: TD3DXHandle;    // Technique handle for RenderHDREnvMap
    m_hTechRenderLight: TD3DXHandle;        // Technique handle for RenderLight
    m_hWorldView: TD3DXHandle;              // Handle for world+view matrix in effect
    m_hProj: TD3DXHandle;                   // Handle for projection matrix in effect
    m_htxCubeMap: TD3DXHandle;              // Handle for the cube texture in effect
    m_htxScene: TD3DXHandle;                // Handle for the scene texture in effect
    m_hLightIntensity: TD3DXHandle;         // Handle for the light intensity in effect
    m_hLightPosView: TD3DXHandle;           // Handle for view space light positions in effect
    m_hReflectivity: TD3DXHandle;           // Handle for reflectivity in effect

    m_fWorldRotX: Single;                   // World rotation state X-axis
    m_fWorldRotY: Single;                   // World rotation state Y-axis
    m_bShowHelp: Boolean;
    m_bUseFloatCubeMap: Boolean;            // Whether we use floating point format cube map
    m_fReflectivity: Single;                // Reflectivity value. Ranges from 0 to 1.

  protected
    function OneTimeSceneInit: HRESULT; override;
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function FinalCleanup: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

    procedure ResetParameters;
    function RenderText: HRESULT;

    function LoadMesh(tszName: PChar; out Mesh: CD3DMesh): HRESULT;

    function RenderSceneIntoCubeMap: HRESULT;
    function RenderScene(const pmView, pmProj: TD3DXMatrix; bRenderEnvMappedMesh: Boolean): HRESULT;

  public
    constructor Create;
    destructor Destroy; override;
    function MsgProc(hWnd: HWND; uMsg: LongWord; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;


//-----------------------------------------------------------------------------
// Global access to the app (needed for the global WndProc())
//-----------------------------------------------------------------------------
var
  g_pApp: CMyD3DApplication = nil;
  g_hInst: THandle {HINSTANCE} = 0;


implementation

uses
  SysUtils, Math;

const
  UnitName = 'HDRCubeMapUnit.pas';


function ComputeBoundingSphere(Mesh: CD3DMesh; out pvCtr: TD3DXVector3; out pfRadius: Single): HRESULT;
var
  pMeshObj: ID3DXMesh;
  decl: TFVFDeclaration;
  pVB: Pointer;
  uStride: LongWord;
begin
  // Obtain the bounding sphere
  pMeshObj := Mesh.GetSysMemMesh;
  if (pMeshObj = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Obtain the declaration
  Result := pMeshObj.GetDeclaration(decl);
  if FAILED(Result) then Exit;

  // Lock the vertex buffer
  Result := pMeshObj.LockVertexBuffer(D3DLOCK_READONLY, pVB);
  if FAILED(Result) then Exit;

  // Compute the bounding sphere
  uStride := D3DXGetDeclVertexSize(@decl, 0);
  Result := D3DXComputeBoundingSphere(PD3DXVector3(pVB), pMeshObj.GetNumVertices,
                                      uStride, pvCtr, pfRadius);
  pMeshObj.UnlockVertexBuffer;
end;


{ CObj }

constructor CObj.Create;
begin
  D3DXMatrixIdentity(m_mWorld);
  m_Mesh:= CD3DMesh.Create; // Mesh object
end;

destructor CObj.Destroy;
begin
  m_Mesh.Free;
  inherited;
end;

function CObj.WorldCenterAndScaleToRadius(fRadius: Single): HRESULT;
var
  fRadiusBound: Single;
  vCtr: TD3DXVector3;
  mScale: TD3DXMatrixA16;
begin
  //
  // Compute the world transformation matrix
  // to center the mesh at origin in world space
  // and scale its size to the specified radius.
  //
  Result := ComputeBoundingSphere(m_Mesh, vCtr, fRadiusBound);
  if FAILED(Result) then Exit;

  D3DXMatrixTranslation(m_mWorld, -vCtr.x, -vCtr.y, -vCtr.z);
  D3DXMatrixScaling(mScale, fRadius / fRadiusBound,
                            fRadius / fRadiusBound,
                            fRadius / fRadiusBound);
  D3DXMatrixMultiply(m_mWorld, m_mWorld, mScale);
end; // WorldCenterAndScaleToRadius




{ COrbit }

constructor COrbit.Create;
begin
  inherited;
  m_vAxis:= D3DXVector3(0.0, 1.0, 0.0);
  m_fRadius:= 1.0;
  m_fSpeed:= D3DX_PI;
end;

procedure COrbit.SetOrbit(const vAxis: TD3DXVector3; fRadius, fSpeed: Single);
var
  m: TD3DXMatrixA16;
  vX: TD3DXVector3;
  vRot: TD3DXVector3;
  fAng: Single;
  vXxRot: TD3DXVector3;  // X cross vRot
begin
  // Call this after m_mWorld is initialized
  m_vAxis := vAxis; m_fRadius := fRadius; m_fSpeed := fSpeed;
  D3DXVec3Normalize(m_vAxis, m_vAxis);

  // Translate by m_fRadius in -Y direction
  D3DXMatrixTranslation(m, 0.0, -m_fRadius, 0.0);
  D3DXMatrixMultiply(m_mWorld, m_mWorld, m);

  // Apply rotation from X axis to the orbit axis
  vX:= D3DXVector3(1.0, 0.0, 0.0);
  D3DXVec3Cross(vRot, m_vAxis, vX);  // Axis of rotation
  // If the cross product is 0, m_vAxis is on the X axis
  // So we either rotate 0 or PI.
  if (D3DXVec3LengthSq(vRot) = 0) then
  begin
    if (m_vAxis.x < 0.0)
      then D3DXMatrixRotationY(m, D3DX_PI) // PI
      else D3DXMatrixIdentity(m);          // 0
  end else
  begin
    fAng := ArcCos(D3DXVec3Dot(m_vAxis, vX));  // Angle to rotate
    // Find out direction to rotate
    D3DXVec3Cross(vXxRot, vX, vRot);
    if (D3DXVec3Dot(vXxRot, m_vAxis) >= 0) then fAng := -fAng;
    D3DXMatrixRotationAxis(m, vRot, fAng);
  end;
  D3DXMatrixMultiply(m_mWorld, m_mWorld, m);
end;

procedure COrbit.Orbit(fElapsedTime: Single);
var
  m: TD3DXMatrixA16;
begin
  // Compute the orbit transformation and apply to m_mWorld
  D3DXMatrixRotationAxis(m, m_vAxis, m_fSpeed * fElapsedTime);
  D3DXMatrixMultiply(m_mWorld, m_mWorld, m);
end;



{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor.   Paired with ~CMyD3DApplication()
//       Member variables should be initialized to a known state here.
//       The application window has not yet been created and no Direct3D device
//       has been created, so any initialization that depends on a window or
//       Direct3D should be deferred to a later stage.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
var
  i: Integer;
begin
  inherited Create;
  m_Camera:= CModelViewerCamera.Create;
  for i:= Low(m_EnvMesh) to High(m_EnvMesh) do m_EnvMesh[i]:= CObj.Create;
  for i:= Low(m_Orbit) to High(m_Orbit) do m_Orbit[i]:= COrbit.Create;

  m_strWindowTitle            := 'HDRCubeMap';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bStartFullscreen			    := False;
  m_bShowCursorWhenFullscreen	:= True;

  m_pFont                     := nil;
  m_pFontSmall                := nil;

  m_dwCreationWidth           := 640;
  m_dwCreationHeight          := 480;

  m_fWorldRotX                := 0.0;
  m_fWorldRotY                := 0.0;

  m_bShowHelp                 := True;

  m_pEffect                   := nil;
  m_pRoom                     := nil;
  m_pLightMesh                := nil;
  m_pVertDecl                 := nil;

  m_hTechRenderScene          := nil;
  m_hTechRenderHDREnvMap      := nil;
  m_hTechRenderLight          := nil;
  m_hWorldView                := nil;
  m_hProj                     := nil;
  m_htxCubeMap                := nil;
  m_htxScene                  := nil;
  m_hLightIntensity           := nil;
  m_hLightPosView             := nil;
  m_hReflectivity             := nil;

  m_pCubeMapFp                := nil;
  m_pCubeMap32                := nil;
  m_pDepthCube                := nil;

  m_nCurrMesh                 := 0;
  ResetParameters;
end;


//-----------------------------------------------------------------------------
// Name: ~CMyD3DApplication()
// Desc: Application destructor.  Paired with CMyD3DApplication()
//-----------------------------------------------------------------------------
destructor CMyD3DApplication.Destroy;
var
  i: Integer;
begin
  for i:= Low(m_Orbit) to High(m_Orbit) do FreeAndNil(m_Orbit[i]);
  for i:= Low(m_EnvMesh) to High(m_EnvMesh) do FreeAndNil(m_EnvMesh[i]);
  FreeAndNil(m_Camera);
  inherited;
end;


//-----------------------------------------------------------------------------
// Name: ResetParameters()
// Desc: Reset light and material parameters to default values
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.ResetParameters;
begin
  m_bUseFloatCubeMap := True;
  m_fReflectivity    := 0.4;
  m_vLightIntensity  := D3DXVector4(24.0, 24.0, 24.0, 24.0);

  if (m_pEffect <> nil) then
  begin
    m_pEffect.SetFloat(m_hReflectivity, m_fReflectivity);
    m_pEffect.SetVector(m_hLightIntensity, m_vLightIntensity);
  end;
end;


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Paired with FinalCleanup().
//       The window has been created and the IDirect3D9 interface has been
//       created, but the device has not been created yet.  Here you can
//       perform application-related initialization and cleanup that does
//       not depend on a device.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HRESULT;
begin
  // Initialize camera parameters
  m_Camera.SetModelCenter(D3DXVector3Zero);
  m_Camera.SetRadius(3.0);
  m_Camera.SetEnablePositionMovement(False);

  try
    m_pRoom := CD3DMesh.Create;
    m_pLightMesh := CD3DMesh.Create;
  except
    Result:= E_OUTOFMEMORY;
    Exit;
  end;

  // Set the light positions
  {}
    m_aLights[0].vPos := D3DXVector4(-3.5, 2.3, -4.0, 1.0);
    m_aLights[0].vMoveDir := D3DXVector4(0.0, 0.0, 1.0, 0.0);
    m_aLights[0].fMoveDist := 8.0;
  {}
  if (NUM_LIGHTS > 1) then
  begin
    m_aLights[1].vPos := D3DXVector4(3.5, 2.3, 4.0, 1.0);
    m_aLights[1].vMoveDir := D3DXVector4(0.0, 0.0, -1.0, 0.0);
    m_aLights[1].fMoveDist := 8.0;
  end;
  if (NUM_LIGHTS > 2) then
  begin
    m_aLights[2].vPos := D3DXVector4(-3.5, 2.3, 4.0, 1.0);
    m_aLights[2].vMoveDir := D3DXVector4(1.0, 0.0, 0.0, 0.0);
    m_aLights[2].fMoveDist := 7.0;
  end;
  if (NUM_LIGHTS > 3) then
  begin
    m_aLights[3].vPos := D3DXVector4(3.5, 2.3, -4.0, 1.0);
    m_aLights[3].vMoveDir := D3DXVector4(-1.0, 0.0, 0.0, 0.0);
    m_aLights[3].fMoveDist := 7.0;
  end;

  // Drawing loading status message until app finishes loading
  SendMessage(m_hWnd, WM_PAINT, 0, 0);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
  if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
     ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
  if (pCaps.DeviceType <> D3DDEVTYPE_REF) then Exit;
  {$ENDIF}

  // Must support cube textures
  if (pCaps.TextureCaps and D3DPTEXTURECAPS_CUBEMAP = 0) then Exit;

  // Must support vertex shader 1.1 or use software vertex processing
  if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,1)) and
     ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;

  // Must support pixel shader 2.0
  if (pCaps.PixelShaderVersion < D3DPS_VERSION(2,0)) then Exit;

  // need to support D3DFMT_A16B16G16R16F render target
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
                                     adapterFormat, D3DUSAGE_RENDERTARGET,
                                     D3DRTYPE_CUBETEXTURE, D3DFMT_A16B16G16R16F))
  then Exit;

  // need to support D3DFMT_A8R8G8B8 render target
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
                                     adapterFormat, D3DUSAGE_RENDERTARGET,
                                     D3DRTYPE_CUBETEXTURE, D3DFMT_A8R8G8B8))
  then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication::LoadMesh
// Desc: Load mesh from file and convert vertices to our format
//-----------------------------------------------------------------------------
function CMyD3DApplication.LoadMesh(tszName: PChar; out Mesh: CD3DMesh): HRESULT;
begin
  Result := Mesh.Create_(m_pd3dDevice, tszName);
  if FAILED(Result) then Exit;

  Result := Mesh.SetVertexDecl(m_pd3dDevice, @g_aVertDecl);
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Paired with DeleteDeviceObjects()
//       The device has been created.  Resources that are not lost on
//       Reset() can be created here -- resources in D3DPOOL_MANAGED,
//       D3DPOOL_SCRATCH, or D3DPOOL_SYSTEMMEM.  Image surfaces created via
//       CreateImageSurface are never lost and can be created here.  Vertex
//       shaders and pixel shaders can also be created here as they are not
//       lost on Reset().
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
var
  i: Integer;
  hDC: Windows.HDC;
  nHeight: Integer;
  dwShaderFlags: DWORD;
  tszPath: array[0..MAX_PATH-1] of Char;
  vCtr: TD3DXVector3;
  fRadius: Single;
  mWorld, m: TD3DXMatrixA16;
  vAxis: TD3DXVector3;
  mIdent: TD3DXMatrixA16;
begin
  // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
  // Debugging vertex shaders requires either REF or software vertex processing, and debugging
  // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
  // experience in the shader debugger.  It enables source level debugging, prevents instruction
  // reordering, prevents dead code elimination, and forces the compiler to compile against the next
  // higher available software target, which ensures that the unoptimized shaders do not exceed
  // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
  // will be unoptimized and forced into software.  See the DirectX documentation for more information
  // about using the shader debugger.
  dwShaderFlags := 0;
  {$IFDEF DEBUG_VS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
  {$ENDIF}

  // Create the effect
  Result := DXUtil_FindMediaFileCch(tszPath, MAX_PATH, 'HDRCubeMap.fx');
  if FAILED(Result) then Exit;

  Result := D3DXCreateEffectFromFile(m_pd3dDevice, tszPath, nil, nil,
                                     dwShaderFlags, nil, m_pEffect, nil);
  if FAILED(Result) then Exit;

  m_hTechRenderScene     := m_pEffect.GetTechniqueByName('RenderScene');
  m_hTechRenderHDREnvMap := m_pEffect.GetTechniqueByName('RenderHDREnvMap');
  m_hTechRenderLight     := m_pEffect.GetTechniqueByName('RenderLight');
  m_hWorldView           := m_pEffect.GetParameterByName(nil, 'g_mWorldView');
  m_hProj                := m_pEffect.GetParameterByName(nil, 'g_mProj');
  m_htxCubeMap           := m_pEffect.GetParameterByName(nil, 'g_txCubeMap');
  m_htxScene             := m_pEffect.GetParameterByName(nil, 'g_txScene');
  m_hLightIntensity      := m_pEffect.GetParameterByName(nil, 'g_vLightIntensity');
  m_hLightPosView        := m_pEffect.GetParameterByName(nil, 'g_vLightPosView');
  m_hReflectivity        := m_pEffect.GetParameterByName(nil, 'g_fReflectivity');

  // Initialize reflectivity
  m_pEffect.SetFloat(m_hReflectivity, m_fReflectivity);

  // Initialize light intensity
  m_pEffect.SetVector(m_hLightIntensity, m_vLightIntensity);

  // Create vertex declaration
  Result := m_pd3dDevice.CreateVertexDeclaration(@g_aVertDecl, m_pVertDecl);
  if FAILED(Result) then Exit;

  // Init the font
  hDC := GetDC(0);
  try
    nHeight := -MulDiv(12, GetDeviceCaps(hDC, LOGPIXELSY), 72);
    Result := D3DXCreateFont(m_pd3dDevice, nHeight, 0, FW_BOLD, 0, FALSE, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH or FF_DONTCARE,
                             'Arial', m_pFont);
    if FAILED(Result) then Exit;
    nHeight := -MulDiv(9, GetDeviceCaps(hDC, LOGPIXELSY), 72);
    Result := D3DXCreateFont(m_pd3dDevice, nHeight, 0, FW_BOLD, 0, FALSE, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH  or  FF_DONTCARE,
                             'Arial', m_pFontSmall);
    if FAILED(Result) then Exit;
  finally
    ReleaseDC(0, hDC);
  end;

  Result:= D3DAPPERR_MEDIANOTFOUND;

  // Load the mesh
  for i := 0 to NUM_MESHES - 1 do
  begin
    if FAILED(LoadMesh(g_atszMeshFileName[i], m_EnvMesh[i].m_Mesh)) then Exit;
    m_EnvMesh[i].WorldCenterAndScaleToRadius(1.0);  // Scale to radius of 1
  end;
  // Load the room object
  if FAILED(LoadMesh('room.x', m_pRoom)) then Exit;
  // Load the light object
  if FAILED(LoadMesh('sphere0.x', m_pLightMesh)) then Exit;

  // Initialize the world matrix for the lights
  if FAILED(ComputeBoundingSphere(m_pLightMesh, vCtr, fRadius)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;
  D3DXMatrixTranslation(mWorld, -vCtr.x, -vCtr.y, -vCtr.z);
  D3DXMatrixScaling(m, LIGHTMESH_RADIUS / fRadius,
                       LIGHTMESH_RADIUS / fRadius,
                       LIGHTMESH_RADIUS / fRadius);
  D3DXMatrixMultiply(mWorld, mWorld, m);
  for i := 0 to NUM_LIGHTS - 1 do
  begin
    D3DXMatrixTranslation(m, m_aLights[i].vPos.x,
                             m_aLights[i].vPos.y,
                             m_aLights[i].vPos.z );
    D3DXMatrixMultiply(m_aLights[i].mWorld, mWorld, m);
  end;

  // Orbit
  for i := 0 to SizeOf(m_Orbit) div SizeOf(m_Orbit[0]) - 1 do
  begin
    if FAILED(LoadMesh(g_OrbitData[i].tszXFile, m_Orbit[i].m_Mesh)) then Exit;

    m_Orbit[i].WorldCenterAndScaleToRadius(0.7);
    vAxis:= g_OrbitData[i].vAxis;
    m_Orbit[i].SetOrbit(vAxis, g_OrbitData[i].fRadius, g_OrbitData[i].fSpeed);
  end;

  // World transform to identity
  D3DXMatrixIdentity(mIdent);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, mIdent);

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Paired with InvalidateDeviceObjects()
//       The device exists, but may have just been Reset().  Resources in
//       D3DPOOL_DEFAULT and any other device state that persists during
//       rendering should be set here.  Render states, matrices, textures,
//       etc., that don't change during rendering can be set once here to
//       avoid redundant state setting during Render() or FrameMove().
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  i: Integer;
  fAspect: Single;
  vFromPt: TD3DXVector3;
  vLookatPt: TD3DXVector3;
begin
  // Restore effect object
  m_pEffect.OnResetDevice;

  // Set the camera projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  m_Camera.SetProjParams(D3DX_PI/4, fAspect, 0.01, 1000.0);

  // Set up the view parameters for the camera
  vFromPt   := D3DXVector3(0.0, 0.0, -10.0);
  vLookatPt := D3DXVector3Zero;
  m_Camera.SetViewParams(vFromPt, vLookatPt);

  // Set up the window parameter for the camera
  m_Camera.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 1.0);

  // Restore the font
  m_pFont.OnResetDevice;
  m_pFontSmall.OnResetDevice;

  // Restore meshes
  for i := 0 to NUM_MESHES - 1 do
    m_EnvMesh[i].m_Mesh.RestoreDeviceObjects(m_pd3dDevice);
  m_pRoom.RestoreDeviceObjects(m_pd3dDevice);
  m_pLightMesh.RestoreDeviceObjects(m_pd3dDevice);
  for i:= 0 to SizeOf(m_Orbit) div SizeOf(m_Orbit[0]) - 1 do
    m_Orbit[i].m_Mesh.RestoreDeviceObjects(m_pd3dDevice);

  // Create the cube textures
  Result := m_pd3dDevice.CreateCubeTexture(ENVMAPSIZE,
                                           1,
                                           D3DUSAGE_RENDERTARGET,
                                           D3DFMT_A16B16G16R16F,
                                           D3DPOOL_DEFAULT,
                                           m_pCubeMapFp,
                                           nil);
  if FAILED(Result) then Exit;

  Result := m_pd3dDevice.CreateCubeTexture(ENVMAPSIZE,
                                           1,
                                           D3DUSAGE_RENDERTARGET,
                                           D3DFMT_A8R8G8B8,
                                           D3DPOOL_DEFAULT,
                                           m_pCubeMap32,
                                           nil);
  if FAILED(Result) then Exit;

  // Create the stencil buffer to be used with the cube textures
  Result := m_pd3dDevice.CreateDepthStencilSurface(ENVMAPSIZE,
                                                   ENVMAPSIZE,
                                                   m_d3dSettings.DepthStencilBufferFormat,
                                                   m_d3dSettings.MultisampleType,
                                                   m_d3dSettings.MultisampleQuality,
                                                   True,
                                                   m_pDepthCube,
                                                   nil);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  i: Integer;
begin
  for i := 0 to SizeOf(m_Orbit) div SizeOf(m_Orbit[0]) - 1 do
    m_Orbit[i].Orbit(m_fElapsedTime);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderSceneIntoCubeMap()
// Desc: Set up the cube map by rendering the scene into it.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderSceneIntoCubeMap: HRESULT;
var
  pCubeMap: IDirect3DCubeTexture9;
  mProj: TD3DXMatrixA16;
  mViewDir: TD3DXMatrixA16;
  pRTOld: IDirect3DSurface9;
  pDSOld: IDirect3DSurface9;
  nFace: TD3DCubemapFaces;
  pSurf: IDirect3DSurface9;
  mView: TD3DXMatrixA16;
begin
  if m_bUseFloatCubeMap then pCubeMap := m_pCubeMapFp else pCubeMap := m_pCubeMap32;

  // The projection matrix has a FOV of 90 degrees and asp ratio of 1
  D3DXMatrixPerspectiveFovLH(mProj, D3DX_PI * 0.5, 1.0, 0.01, 100.0);

  mViewDir:= m_Camera.GetViewMatrix^;
  mViewDir._41 := 0.0; mViewDir._42 := 0.0; mViewDir._43 := 0.0;

  m_pd3dDevice.GetRenderTarget(0, pRTOld);
  if SUCCEEDED(m_pd3dDevice.GetDepthStencilSurface(pDSOld)) then
  begin
    // If the device has a depth-stencil buffer, use
    // the depth stencil buffer created for the cube textures.
    m_pd3dDevice.SetDepthStencilSurface(m_pDepthCube);
  end;

  for nFace := Low(TD3DCubemapFaces) to High(TD3DCubemapFaces) do
  begin
    Result := pCubeMap.GetCubeMapSurface(nFace, 0, pSurf);
    if FAILED(Result) then Continue;

    Result := m_pd3dDevice.SetRenderTarget(0, pSurf);
    SAFE_RELEASE(pSurf);
    if FAILED(Result) then Continue;

    mView := D3DUtil_GetCubeMapViewMatrix(nFace);
    D3DXMatrixMultiply(mView, mViewDir, mView);

    m_pd3dDevice.Clear(0, nil, D3DCLEAR_ZBUFFER, $000000ff, 1.0, 0);

    // Begin the scene
    if SUCCEEDED(m_pd3dDevice.BeginScene) then
    begin
      RenderScene(mView, mProj, False);

      // End the scene.
      m_pd3dDevice.EndScene;
    end;
  end;

  // Restore depth-stencil buffer and render target
  if (pDSOld <> nil) then 
  begin
    m_pd3dDevice.SetDepthStencilSurface(pDSOld);
    SAFE_RELEASE(pDSOld);
  end;
  m_pd3dDevice.SetRenderTarget(0, pRTOld);
  SAFE_RELEASE(pRTOld);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc: Renders the scene with a specific view and projection matrix.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderScene(const pmView, pmProj: TD3DXMatrix; bRenderEnvMappedMesh: Boolean): HRESULT;
 function fmod(x, y: Single): Single;
 begin // Clootie: This is just a hack in absence of real 'fmod' RTL function
   Result:= x / y;
   Result:= x - Trunc(Result) * y;
 end;
var
  p, cPass: LongWord;
  mWorldView: TD3DXMatrixA16;
  avLightPosView: array[0..NUM_LIGHTS-1] of TD3DXVector4;
  i: Integer;
  fDisp: Single;
  vMove: TD3DXVector4;
  pCubeMap: IDirect3DCubeTexture9;
  pMesh: ID3DXMesh;
  pMeshObj: ID3DXMesh;
  m: DWORD;
begin
  m_pEffect.SetMatrix(m_hProj, pmProj);

  // Write camera-space light positions to effect
  for i := 0 to NUM_LIGHTS - 1 do
  begin
    // Animate the lights
    fDisp := (1.0 + Cos(fmod(m_fTime, D3DX_PI))) * 0.5 * m_aLights[i].fMoveDist; // Distance to move
    // vMove := m_aLights[i].vMoveDir * fDisp;  // In vector form
    D3DXVec4Scale(vMove, m_aLights[i].vMoveDir, fDisp);
    D3DXMatrixTranslation(m_aLights[i].mWorking, vMove.x, vMove.y, vMove.z); // Matrix form
    D3DXMatrixMultiply(m_aLights[i].mWorking, m_aLights[i].mWorld, m_aLights[i].mWorking);
    // vMove +:= m_aLights[i].vPos;  // Animated world coordinates
    D3DXVec4Add(vMove, vMove, m_aLights[i].vPos);
    D3DXVec4Transform(avLightPosView[i], vMove, pmView);
  end;
  m_pEffect.SetVectorArray(m_hLightPosView, @avLightPosView, NUM_LIGHTS);

  //
  // Render the environment-mapped mesh if specified
  //

  if (bRenderEnvMappedMesh) then
  begin
    m_pEffect.SetTechnique(m_hTechRenderHDREnvMap);

    // Combine the offset and scaling transformation with
    // rotation from the camera to form the final
    // world transformation matrix.
    D3DXMatrixMultiply(mWorldView, m_EnvMesh[m_nCurrMesh].m_mWorld, m_Camera.GetWorldMatrix^);
    D3DXMatrixMultiply(mWorldView, mWorldView, pmView);
    m_pEffect.SetMatrix(m_hWorldView, mWorldView);

    if m_bUseFloatCubeMap then pCubeMap := m_pCubeMapFp else pCubeMap := m_pCubeMap32;
    m_pEffect.SetTexture(m_htxCubeMap, pCubeMap);

    m_pEffect._Begin(@cPass, D3DXFX_DONOTSAVESTATE);
    for p := 0 to cPass - 1 do
    begin
      m_pEffect.Pass( p);
      pMesh := m_EnvMesh[m_nCurrMesh].m_Mesh.GetLocalMesh;
      for i := 0 to m_EnvMesh[m_nCurrMesh].m_Mesh.m_dwNumMaterials - 1 do
        pMesh.DrawSubset(i);
    end;
    m_pEffect._End;
  end;

  //
  // Render light spheres
  //

  m_pEffect.SetTechnique(m_hTechRenderLight);

  m_pEffect._Begin(@cPass, D3DXFX_DONOTSAVESTATE);
  for p := 0 to cPass - 1 do
  begin
    m_pEffect.Pass(p);

    for i := 0 to NUM_LIGHTS - 1 do
    begin
      D3DXMatrixMultiply(mWorldView, m_aLights[i].mWorking, pmView);
      m_pEffect.SetMatrix(m_hWorldView, mWorldView);
      m_pLightMesh.Render(m_pd3dDevice);
    end;
  end;
  m_pEffect._End;

  //
  // Render the rest of the scene
  //

  m_pEffect.SetTechnique(m_hTechRenderScene);

  m_pEffect._Begin(@cPass, D3DXFX_DONOTSAVESTATE);
  for p := 0 to cPass - 1 do
  begin
     m_pEffect.Pass(p);

     //
     // Orbits
     //
     for i := Low(m_Orbit) to High(m_Orbit) do
     begin
       D3DXMatrixMultiply(mWorldView, m_Orbit[i].m_mWorld, pmView);
       m_pEffect.SetMatrix(m_hWorldView, mWorldView);
       // Obtain the D3DX mesh object
       pMeshObj := m_Orbit[i].m_Mesh.GetLocalMesh;
       // Iterate through each subset and render with its texture
       for m := 0 to m_Orbit[i].m_Mesh.m_dwNumMaterials - 1 do
       begin
         m_pEffect.SetTexture(m_htxScene, m_Orbit[i].m_Mesh.m_pTextures^[m]);
         pMeshObj.DrawSubset(m);
       end;
     end;

     //
     // The room object (walls, floor, ceiling)
     //

     m_pEffect.SetMatrix(m_hWorldView, pmView);
     pMeshObj := m_pRoom.GetLocalMesh;
     // Iterate through each subset and render with its texture
     for m := 0 to m_pRoom.m_dwNumMaterials - 1 do
     begin
       m_pEffect.SetTexture(m_htxScene, m_pRoom.m_pTextures^[m]);
       pMeshObj.DrawSubset(m);
     end;
  end;
  m_pEffect._End;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
begin
  // Call this in Render() instead of FrameMove() so we can
  // still move camera while single stepping.
  m_Camera.FrameMove(m_fElapsedTime);

  RenderSceneIntoCubeMap;

  // Clear the render buffers
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    RenderScene(m_Camera.GetViewMatrix^, m_Camera.GetProjMatrix^, True);

    // Render stats and help text
    RenderText;

    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Draw the help & statistics for running sample
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderText: HRESULT;
var
  rc: TRect;
  S: AnsiString;
begin
{$IFDEF UNICODE}
  For UNICODE you should use "m_pFont.DrawTextW" routine!!!
{$ENDIF}

  // Output statistics
  SetRect(rc, 2, 0, 0, 0);
  m_pFont.DrawTextA(nil, m_strFrameStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));
  Inc(rc.top, 20);
  m_pFont.DrawTextA(nil, m_strDeviceStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));

  // Output statistics & help
  if m_bShowHelp then
  begin
    SetRect(rc, 2, 40, 0, 0);
    m_pFontSmall.DrawTextA(nil, 'Press F1 for help',
                          -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));

    SetRect(rc, 10, m_d3dsdBackBuffer.Height - 188, 0, 0);
    m_pFontSmall.DrawTextA(nil, 'Keyboard controls:', -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));
    Inc(rc.top, 20);
    m_pFontSmall.DrawTextA(nil,
                           'Rotate object'#10'Adjust camera'#10'Zoom In/Out'#10 +
                           'Change mesh'#10'Change cube map format'#10 +
                           'Adjust reflectivity'#10'Adjust light intensity'#10 +
                           'Reset parameters'#10 +
                           'Help'#10'Change device'#10'Exit',
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));
    SetRect(rc, 168, m_d3dsdBackBuffer.Height - 168, 0, 0);
    m_pFontSmall.DrawTextA(nil,
                           'Left drag mouse'#10'Right drag mouse'#10'Mouse wheel'#10 +
                           'N'#10'F'#10 +
                           'E,D'#10'W,S'#10 +
                           'R'#10 +
                           'F1'#10'F2'#10'Esc',
                            -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));
  end else
  begin
    SetRect(rc, 2, 40, 0, 0);
    m_pFontSmall.DrawTextA(nil, 'Press F1 for help',
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));
  end;

  //
  // Output cube mapping stat
  //
  SetRect(rc, 270, m_d3dsdBackBuffer.Height - 48, 0, 0);
  m_pFontSmall.DrawTextA(nil,
                         'Cube map format ( F )'#10 +
                         'Material reflectivity ( e/E, d/D )'#10 +
                         'Light intensity ( w/W, s/S )',
                         -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));
  SetRect(rc, 450, m_d3dsdBackBuffer.Height - 48, 0, 0);


  if m_bUseFloatCubeMap then S:= 'A16B16G16R16F ( D3D9 / HDR )' else S:= 'A8R8G8B8 ( D3D8 )';
  S:= Format('%s'#10'%.2f'#10'%.1f', [S, m_fReflectivity, m_vLightIntensity.x]);
  m_pFontSmall.DrawTextA(nil,
                         PAnsiChar(S),
                         -1, @rc, DT_NOCLIP, D3DXColorToDWord(HELPTEXTCOLOR));

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: LongWord;
  wParam: WPARAM; lParam: LPARAM): LRESULT;
begin
  m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  case uMsg of
    //
    // Use WM_CHAR to handle parameter adjustment so
    // that we can control the granularity based on
    // the letter cases.
    WM_CHAR:
    begin
      case Char(wParam) of
        'W', 'w':
        begin
          if ('w' = Char(wParam))
            then D3DXVec4Add(m_vLightIntensity, m_vLightIntensity, D3DXVector4(0.1, 0.1, 0.1, 0.0))
            else D3DXVec4Add(m_vLightIntensity, m_vLightIntensity, D3DXVector4(10.0, 10.0, 10.0, 0.0));
          m_pEffect.SetVector(m_hLightIntensity, m_vLightIntensity);
        end;

        'S', 's':
        begin
          if ('s' = Char(wParam))
            then D3DXVec4Subtract(m_vLightIntensity, m_vLightIntensity, D3DXVector4(0.1, 0.1, 0.1, 0.0))
            else D3DXVec4Subtract(m_vLightIntensity, m_vLightIntensity, D3DXVector4(10.0, 10.0, 10.0, 0.0));

          if (m_vLightIntensity.x < 0.0) then
          begin
            m_vLightIntensity.x := 0.0;
            m_vLightIntensity.y := 0.0;
            m_vLightIntensity.z := 0.0;
          end;
          m_pEffect.SetVector(m_hLightIntensity, m_vLightIntensity);
        end;

        'D', 'd':
        begin
          if ('d' = Char(wParam))
            then m_fReflectivity := m_fReflectivity - 0.01
            else m_fReflectivity := m_fReflectivity - 0.1;
          if (m_fReflectivity < 0) then m_fReflectivity := 0;
          m_pEffect.SetFloat( m_hReflectivity, m_fReflectivity);
        end;

        'E', 'e':
        begin
          if ('e' = Char(wParam))
            then m_fReflectivity := m_fReflectivity + 0.01
            else m_fReflectivity := m_fReflectivity + 0.1;
          if (m_fReflectivity > 1.0) then m_fReflectivity := 1.0;
          m_pEffect.SetFloat(m_hReflectivity, m_fReflectivity);
        end;
      end;

      Result:= 0;
      Exit;
    end;

    WM_KEYDOWN:
    begin
      case Char(wParam) of
        'N':
        begin
          Inc(m_nCurrMesh);
          if (m_nCurrMesh = NUM_MESHES) then m_nCurrMesh := 0;
        end;

        'F': m_bUseFloatCubeMap := not m_bUseFloatCubeMap;
      end;

      Result:= 0;
      Exit;
    end;

    WM_COMMAND:
    begin
      case LOWORD(wParam) of
        IDM_TOGGLEHELP:   m_bShowHelp := not m_bShowHelp;
        IDM_RESETPARAMS:  ResetParameters;
      end;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.  Paired with RestoreDeviceObjects()
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
var
  i: Integer;
begin
  m_pFont.OnLostDevice;
  m_pFontSmall.OnLostDevice;

  for i := 0 to NUM_MESHES - 1 do
    m_EnvMesh[i].m_Mesh.InvalidateDeviceObjects;
  m_pRoom.InvalidateDeviceObjects;
  m_pLightMesh.InvalidateDeviceObjects;
  for i := 0 to SizeOf(m_Orbit) div SizeOf(m_Orbit[0]) - 1 do
    m_Orbit[i].m_Mesh.InvalidateDeviceObjects;

  m_pEffect.OnLostDevice;

  SAFE_RELEASE(m_pCubeMapFp);
  SAFE_RELEASE(m_pCubeMap32);
  SAFE_RELEASE(m_pDepthCube);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Paired with InitDeviceObjects()
//       Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
var
  i: Integer;
begin
  for i := 0 to NUM_MESHES - 1 do
    m_EnvMesh[i].m_Mesh.Destroy_;
  SAFE_RELEASE(m_pFont);
  SAFE_RELEASE(m_pFontSmall);
  SAFE_RELEASE(m_pVertDecl);

  m_pRoom.Destroy_;
  m_pLightMesh.Destroy_;
  for i := 0 to SizeOf(m_Orbit) div SizeOf(m_Orbit[0]) - 1 do
    m_Orbit[i].m_Mesh.Destroy_;

  SAFE_RELEASE(m_pEffect);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Paired with OneTimeSceneInit()
//       Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  FreeAndNil(m_pRoom);
  FreeAndNil(m_pLightMesh);

  Result:= S_OK;
end;


begin
  // This is initialization 
  g_aVertDecl[3]:= D3DDECL_END;
end.

